// Copyright 2023 Simo Sorce
// See LICENSE.txt file for terms

use std::collections::HashMap;
use std::sync::{RwLock, RwLockReadGuard, RwLockWriteGuard};

use crate::config;
use crate::error::Result;
use crate::interface::*;
use crate::misc::copy_sized_string;
use crate::session::Session;
use crate::token::Token;

static SLOT_DESCRIPTION: [CK_UTF8CHAR; 64usize] =
    *b"Kryoptic Slot                                                   ";
static MANUFACTURER_ID: [CK_UTF8CHAR; 32usize] =
    *b"Kryoptic                        ";

#[derive(Debug)]
pub struct Slot {
    slot_info: CK_SLOT_INFO,
    token: RwLock<Token>,
    sessions: HashMap<CK_SESSION_HANDLE, RwLock<Session>>,
}

impl Slot {
    pub fn new(config: &config::Slot) -> Result<Slot> {
        let dbtype: &str;
        let dbargs: Option<String>;

        match &config.dbtype {
            Some(t) => dbtype = t.as_str(),
            None => return Err(CKR_GENERAL_ERROR)?,
        }
        match &config.dbargs {
            Some(p) => dbargs = Some(p.clone()),
            None => dbargs = None,
        }

        let mut slot = Slot {
            slot_info: CK_SLOT_INFO {
                slotDescription: [0; 64],
                manufacturerID: [0; 32],
                flags: CKF_TOKEN_PRESENT,
                hardwareVersion: CK_VERSION { major: 0, minor: 0 },
                firmwareVersion: CK_VERSION { major: 0, minor: 0 },
            },
            token: RwLock::new(Token::new(dbtype, dbargs)?),
            sessions: HashMap::new(),
        };

        /* fill strings */
        copy_sized_string(
            match &config.description {
                Some(d) => d.as_bytes(),
                None => &SLOT_DESCRIPTION,
            },
            &mut slot.slot_info.slotDescription,
        );
        copy_sized_string(
            match &config.manufacturer {
                Some(m) => m.as_bytes(),
                None => &MANUFACTURER_ID,
            },
            &mut slot.slot_info.manufacturerID,
        );
        Ok(slot)
    }

    pub fn get_slot_info(&self) -> &CK_SLOT_INFO {
        &self.slot_info
    }

    pub fn get_token_info(&self) -> CK_TOKEN_INFO {
        let tok = self.token.read().unwrap();
        *tok.get_token_info()
    }

    pub fn get_token(&self) -> Result<RwLockReadGuard<'_, Token>> {
        match self.token.read() {
            Ok(token) => {
                if token.is_initialized() {
                    Ok(token)
                } else {
                    /* FIXME: once we have CKR_TOKEN_NOT_INITIALIZED as an
                     * available error, we should rreturn that instead */
                    Err(KRR_TOKEN_NOT_INITIALIZED)?
                }
            }
            Err(_) => Err(CKR_GENERAL_ERROR)?,
        }
    }

    pub fn get_token_mut(
        &self,
        nochecks: bool,
    ) -> Result<RwLockWriteGuard<'_, Token>> {
        match self.token.write() {
            Ok(token) => {
                if nochecks {
                    Ok(token)
                } else if token.is_initialized() {
                    Ok(token)
                } else {
                    /* FIXME: once we have CKR_TOKEN_NOT_INITIALIZED as an
                     * available error, we should rreturn that instead */
                    Err(KRR_TOKEN_NOT_INITIALIZED)?
                }
            }
            Err(_) => Err(CKR_GENERAL_ERROR)?,
        }
    }

    pub fn add_session(&mut self, handle: CK_SESSION_HANDLE, session: Session) {
        self.sessions.insert(handle, RwLock::new(session));
    }

    pub fn get_session(
        &self,
        handle: CK_SESSION_HANDLE,
    ) -> Result<RwLockReadGuard<'_, Session>> {
        match self.sessions.get(&handle) {
            Some(s) => match s.read() {
                Ok(sess) => Ok(sess),
                Err(_) => Err(CKR_GENERAL_ERROR)?,
            },
            None => Err(CKR_SESSION_HANDLE_INVALID)?,
        }
    }

    pub fn get_session_mut(
        &self,
        handle: CK_SESSION_HANDLE,
    ) -> Result<RwLockWriteGuard<'_, Session>> {
        match self.sessions.get(&handle) {
            Some(s) => match s.write() {
                Ok(sess) => Ok(sess),
                Err(_) => Err(CKR_GENERAL_ERROR)?,
            },
            None => Err(CKR_SESSION_HANDLE_INVALID)?,
        }
    }

    pub fn has_sessions(&self) -> bool {
        self.sessions.len() > 0
    }

    pub fn has_ro_sessions(&self) -> bool {
        for (_key, val) in self.sessions.iter() {
            match val.read().unwrap().get_session_info().state {
                CKS_RO_PUBLIC_SESSION | CKS_RO_USER_FUNCTIONS => return true,
                _ => (),
            }
        }
        false
    }

    pub fn change_session_states(&self, user_type: CK_USER_TYPE) -> Result<()> {
        for (_key, val) in self.sessions.iter() {
            let ret = val.write().unwrap().change_session_state(user_type);
            if ret != CKR_OK {
                return Err(ret)?;
            }
        }
        Ok(())
    }

    pub fn invalidate_session_states(&self) {
        for (_key, val) in self.sessions.iter() {
            let _ = val
                .write()
                .unwrap()
                .change_session_state(CK_UNAVAILABLE_INFORMATION);
        }
    }

    pub fn drop_session(&mut self, handle: CK_SESSION_HANDLE) {
        self.sessions.remove(&handle);
    }

    pub fn drop_all_sessions(&mut self) -> Vec<CK_SESSION_HANDLE> {
        let mut handles =
            Vec::<CK_SESSION_HANDLE>::with_capacity(self.sessions.len());
        for key in self.sessions.keys() {
            handles.push(*key);
        }
        self.sessions.clear();
        handles
    }

    pub fn finalize(&mut self) -> Result<()> {
        self.drop_all_sessions();
        self.token.write().unwrap().save()
    }
}
